@@ -47,6 +47,11 @@ void get_fat(FAT_ENTRY * entry, void *fat, uint32_t cluster, DOS_FS * fs)
 {
     unsigned char *ptr;
 
+    if (cluster > fs->data_clusters + 1) {
+	die("Internal error: cluster out of range in get_fat() (%lu > %lu).",
+		(unsigned long)cluster, (unsigned long)(fs->data_clusters + 1));
+    }
+
     switch (fs->fat_bits) {
     case 12:
 	ptr = &((unsigned char *)fat)[cluster * 3 / 2];
@@ -94,7 +99,7 @@ void read_fat(DOS_FS * fs)
     fs->fat = NULL;
     fs->cluster_owner = NULL;
 
-    total_num_clusters = fs->clusters + 2UL;
+    total_num_clusters = fs->data_clusters + 2;
     eff_size = (total_num_clusters * fs->fat_bits + 7) / 8ULL;
 
     if (fs->fat_bits != 12)
@@ -155,17 +160,19 @@ void read_fat(DOS_FS * fs)
     memset(fs->cluster_owner, 0, (total_num_clusters * sizeof(DOS_FILE *)));
 
     /* Truncate any cluster chains that link to something out of range */
-    for (i = 2; i < fs->clusters + 2; i++) {
+    for (i = 2; i < fs->data_clusters + 2; i++) {
 	FAT_ENTRY curEntry;
 	get_fat(&curEntry, fs->fat, i, fs);
 	if (curEntry.value == 1) {
-	    printf("Cluster %ld out of range (1). Setting to EOF.\n", (long)(i - 2));
+	    printf("Cluster %ld out of range (1). Setting to EOF.\n",
+		   (long)(i - 2));
 	    set_fat(fs, i, -1);
 	}
-	if (curEntry.value >= fs->clusters + 2 &&
+	if (curEntry.value >= fs->data_clusters + 2 &&
 	    (curEntry.value < FAT_MIN_BAD(fs))) {
 	    printf("Cluster %ld out of range (%ld > %ld). Setting to EOF.\n",
-		   (long)(i - 2), (long)curEntry.value, (long)(fs->clusters + 2 - 1));
+		   (long)(i - 2), (long)curEntry.value,
+		   (long)(fs->data_clusters + 2 - 1));
 	    set_fat(fs, i, -1);
 	}
     }
@@ -188,12 +195,22 @@ void set_fat(DOS_FS * fs, uint32_t cluster, int32_t new)
 {
     unsigned char *data = NULL;
     int size;
-    loff_t offs;
+    off_t offs;
+
+    if (cluster > fs->data_clusters + 1) {
+	die("Internal error: cluster out of range in set_fat() (%lu > %lu).",
+		(unsigned long)cluster, (unsigned long)(fs->data_clusters + 1));
+    }
 
     if (new == -1)
 	new = FAT_EOF(fs);
     else if ((long)new == -2)
 	new = FAT_BAD(fs);
+    else if (new > fs->data_clusters + 1) {
+	die("Internal error: new cluster out of range in set_fat() (%lu > %lu).",
+		(unsigned long)new, (unsigned long)(fs->data_clusters + 1));
+    }
+
     switch (fs->fat_bits) {
     case 12:
 	data = fs->fat + cluster * 3 / 2;
@@ -205,10 +222,12 @@ void set_fat(DOS_FS * fs, uint32_t cluster, int32_t new)
 	    data[1] = new >> 4;
 	} else {
 	    FAT_ENTRY subseqEntry;
-	    get_fat(&subseqEntry, fs->fat, cluster + 1, fs);
+	    if (cluster != fs->data_clusters + 1)
+		get_fat(&subseqEntry, fs->fat, cluster + 1, fs);
+	    else
+		subseqEntry.value = 0;
 	    data[0] = new & 0xff;
-	    data[1] = (new >> 8) | (cluster == fs->clusters - 1 ? 0 :
-				    (0xff & subseqEntry.value) << 4);
+	    data[1] = (new >> 8) | ((0xff & subseqEntry.value) << 4);
 	}
 	size = 2;
 	break;
@@ -272,10 +291,9 @@ uint32_t next_cluster(DOS_FS * fs, uint32_t cluster)
     return FAT_IS_EOF(fs, value) ? -1 : value;
 }
 
-loff_t cluster_start(DOS_FS * fs, uint32_t cluster)
+off_t cluster_start(DOS_FS * fs, uint32_t cluster)
 {
-    return fs->data_start + ((loff_t) cluster -
-			     2) * (uint64_t)fs->cluster_size;
+    return fs->data_start + ((off_t)cluster - 2) * (uint64_t)fs->cluster_size;
 }
 
 /**
@@ -312,7 +330,7 @@ void fix_bad(DOS_FS * fs)
 
     if (verbose)
 	printf("Checking for bad clusters.\n");
-    for (i = 2; i < fs->clusters + 2; i++) {
+    for (i = 2; i < fs->data_clusters + 2; i++) {
 	FAT_ENTRY curEntry;
 	get_fat(&curEntry, fs->fat, i, fs);
 
@@ -332,7 +350,7 @@ void reclaim_free(DOS_FS * fs)
     if (verbose)
 	printf("Checking for unused clusters.\n");
     reclaimed = 0;
-    for (i = 2; i < fs->clusters + 2; i++) {
+    for (i = 2; i < fs->data_clusters + 2; i++) {
 	FAT_ENTRY curEntry;
 	get_fat(&curEntry, fs->fat, i, fs);
 
@@ -367,7 +385,7 @@ static void tag_free(DOS_FS * fs, DOS_FILE * owner, uint32_t *num_refs,
     if (start_cluster == 0)
 	start_cluster = 2;
 
-    for (i = start_cluster; i < fs->clusters + 2; i++) {
+    for (i = start_cluster; i < fs->data_clusters + 2; i++) {
 	FAT_ENTRY curEntry;
 	get_fat(&curEntry, fs->fat, i, fs);
 
@@ -418,7 +436,7 @@ void reclaim_file(DOS_FS * fs)
     if (verbose)
 	printf("Reclaiming unconnected clusters.\n");
 
-    total_num_clusters = fs->clusters + 2UL;
+    total_num_clusters = fs->data_clusters + 2;
     num_refs = alloc(total_num_clusters * sizeof(uint32_t));
     memset(num_refs, 0, (total_num_clusters * sizeof(uint32_t)));
 
@@ -431,7 +449,7 @@ void reclaim_file(DOS_FS * fs)
 	get_fat(&curEntry, fs->fat, i, fs);
 
 	next = curEntry.value;
-	if (!get_owner(fs, i) && next && next < fs->clusters + 2) {
+	if (!get_owner(fs, i) && next && next < fs->data_clusters + 2) {
 	    /* Cluster is linked, but not owned (orphan) */
 	    FAT_ENTRY nextEntry;
 	    get_fat(&nextEntry, fs->fat, next, fs);
@@ -483,7 +501,7 @@ void reclaim_file(DOS_FS * fs)
 	/* If this cluster is the head of an orphan chain... */
 	if (get_owner(fs, i) == &orphan && !num_refs[i]) {
 	    DIR_ENT de;
-	    loff_t offset;
+	    off_t offset;
 	    files++;
 	    offset = alloc_rootdir_entry(fs, &de, "FSCK%04dREC");
 	    de.start = htole16(i & 0xffff);
@@ -511,7 +529,7 @@ uint32_t update_free(DOS_FS * fs)
     uint32_t free = 0;
     int do_set = 0;
 
-    for (i = 2; i < fs->clusters + 2; i++) {
+    for (i = 2; i < fs->data_clusters + 2; i++) {
 	FAT_ENTRY curEntry;
 	get_fat(&curEntry, fs->fat, i, fs);
 
